@@ -1,6 +1,4 @@
 /*
- * "$Id$"
- *
  *   Common text filter routines for CUPS.
  *
  *   Copyright 2007-2011 by Apple Inc.
@@ -8,11 +6,8 @@
  *
  *   These coded instructions, statements, and computer programs are the
  *   property of Apple Inc. and are protected by Federal copyright
- *   law.  Distribution and use rights are outlined in the file "LICENSE.txt"
- *   which should have been included with this file.  If this file is
- *   file is missing or damaged, see the license at "http://www.cups.org/".
- *
- *   This file is subject to the Apple OS-Developed Software exception.
+ *   law.  Distribution and use rights are outlined in the file "COPYING"
+ *   which should have been included with this file.
  *
  * Contents:
  *
@@ -26,6 +21,7 @@
  */
 
 #include "textcommon.h"
+#include <limits.h>
 
 
 /*
@@ -485,6 +481,7 @@ TextMain(const char *name,	/* I - Name of filter */
   FILE		*fp;		/* Print file */
   ppd_file_t	*ppd;		/* PPD file */
   int		i,		/* Looping var */
+		empty,		/* Is the input empty? */
 		ch,		/* Current char from file */
 		lastch,		/* Previous char from file */
 		attr,		/* Current attribute */
@@ -497,8 +494,10 @@ TextMain(const char *name,	/* I - Name of filter */
   char		keyword[64],	/* Keyword string */
 		*keyptr;	/* Pointer into string */
   int		keycol;		/* Column where keyword starts */
-  int		ccomment;	/* Inside a C-style comment? */
-  int		cstring;	/* Inside a C string */
+  enum	{NLstyl=-1, NoCmnt, SNTXstyl}
+  		cmntState;	/* Inside a comment */
+  enum	{StrBeg=-1, NoStr, StrEnd}	
+  		strState;	/* Inside a dbl-quoted string */
 
 
  /*
@@ -608,6 +607,8 @@ TextMain(const char *name,	/* I - Name of filter */
 
     if (PageColumns < 1)
     {
+      if (fp != stdin)
+        fclose(fp);
       fprintf(stderr, "ERROR: Bad columns value %d.\n", PageColumns);
       return (1);
     }
@@ -619,6 +620,8 @@ TextMain(const char *name,	/* I - Name of filter */
 
     if (CharsPerInch <= 0.0)
     {
+      if (fp != stdin)
+        fclose(fp);
       fprintf(stderr, "ERROR: Bad cpi value %f.\n", CharsPerInch);
       return (1);
     }
@@ -630,6 +633,8 @@ TextMain(const char *name,	/* I - Name of filter */
 
     if (LinesPerInch <= 0.0)
     {
+      if (fp != stdin)
+        fclose(fp);
       fprintf(stderr, "ERROR: Bad lpi value %f.", LinesPerInch);
       return (1);
     }
@@ -638,15 +643,52 @@ TextMain(const char *name,	/* I - Name of filter */
   if (PrettyPrint)
     PageTop -= 216.0f / LinesPerInch;
 
-  Copies = atoi(argv[4]);
+ /*
+  * Allocate memory for the page...
+  */
 
-  WriteProlog(argv[3], argv[2], getenv("CLASSIFICATION"),
-              cupsGetOption("page-label", num_options, options), ppd);
+  SizeColumns = (PageRight - PageLeft) / 72.0 * CharsPerInch;
+  SizeLines   = (PageTop - PageBottom) / 72.0 * LinesPerInch;
+
+ /*
+  * Enforce minimum size...
+  */
+  if (SizeColumns < 1)
+    SizeColumns = 1;
+  if (SizeLines < 1)
+    SizeLines = 1;
+
+  if (SizeLines >= INT_MAX / SizeColumns / sizeof(lchar_t))
+  {
+    fprintf(stderr, "ERROR: bad page size\n");
+    exit(1);
+  }
+
+  Page    = calloc(sizeof(lchar_t *), SizeLines);
+  if (!Page)
+  {
+    fprintf(stderr, "ERROR: cannot allocate memory for page\n");
+    exit(1);
+  }
+
+  Page[0] = calloc(sizeof(lchar_t), SizeColumns * SizeLines);
+  if (!Page[0])
+  {
+    free(Page);
+    fprintf(stderr, "ERROR: cannot allocate memory for page\n");
+    exit(1);
+  }
+
+  for (i = 1; i < SizeLines; i ++)
+    Page[i] = Page[0] + i * SizeColumns;
+
+  Copies = atoi(argv[4]);
 
  /*
   * Read text from the specified source and print it...
   */
 
+  empty        = 1;
   lastch       = 0;
   column       = 0;
   line         = 0;
@@ -654,11 +696,19 @@ TextMain(const char *name,	/* I - Name of filter */
   attr         = 0;
   keyptr       = keyword;
   keycol       = 0;
-  ccomment     = 0;
-  cstring      = 0;
+  cmntState     = NoCmnt;
+  strState      = NoStr;
 
   while ((ch = getutf8(fp)) >= 0)
   {
+    if (empty)
+    {
+      /* Found the first valid character, write file header */
+      empty = 0;
+      WriteProlog(argv[3], argv[2], getenv("CLASSIFICATION"),
+		  cupsGetOption("page-label", num_options, options), ppd);
+    }
+
    /*
     * Control codes:
     *
@@ -782,7 +832,10 @@ TextMain(const char *name,	/* I - Name of filter */
           column = 0;
 	  keycol = 0;
 
-          if (!ccomment && !cstring)
+	  if (cmntState == NLstyl)
+	  	cmntState = NoCmnt;
+
+          if (!cmntState && !strState)
 	    attr &= ~(ATTR_ITALIC | ATTR_BOLD | ATTR_RED | ATTR_GREEN | ATTR_BLUE);
 
           if (line >= SizeLines)
@@ -805,7 +858,10 @@ TextMain(const char *name,	/* I - Name of filter */
           keyptr = keyword;
 	  keycol = column;
 
-          if (!ccomment && !cstring)
+	  if (cmntState == NLstyl)
+	  	cmntState = NoCmnt;
+
+          if (!cmntState && !strState)
 	    attr &= ~(ATTR_ITALIC | ATTR_BOLD | ATTR_RED | ATTR_GREEN | ATTR_BLUE);
           break;
 
@@ -837,7 +893,10 @@ TextMain(const char *name,	/* I - Name of filter */
 	  keycol = 0;
           line   = 0;
 
-          if (!ccomment && !cstring)
+	  if (cmntState == NLstyl)
+	  	cmntState = NoCmnt;
+
+          if (!cmntState && !strState)
 	    attr &= ~(ATTR_ITALIC | ATTR_BOLD | ATTR_RED | ATTR_GREEN | ATTR_BLUE);
 
           if (page_column >= PageColumns)
@@ -920,8 +979,7 @@ TextMain(const char *name,	/* I - Name of filter */
 	      *keyptr = '\0';
 	      keyptr  = keyword;
 
-	      if (!(attr & ATTR_ITALIC) &&
-	          bsearch(&keyptr, Keywords, NumKeywords, sizeof(char *),
+	      if (bsearch(&keyptr, Keywords, NumKeywords, sizeof(char *),
 	                  compare_keywords))
               {
 	       /*
@@ -937,62 +995,71 @@ TextMain(const char *name,	/* I - Name of filter */
 		}
 	      }
 	    }
-	    else if ((isalnum(ch & 255) || ch == '_') && !ccomment && !cstring)
+
+	   /*
+	    * Look for Syntax-transition Starts...
+	    */
+	    if (!cmntState && !strState)
 	    {
-	     /*
-	      * Add characters to the current keyword (if they'll fit).
-	      */
+	      if ((isalnum(ch & 255) || ch == '_'))
+	      {
+	       /*
+	        * Add characters to the current keyword (if they'll fit).
+	        */
 
-              if (keyptr == keyword)
-	        keycol = column;
+	        if (keyptr == keyword)
+	          keycol = column;
 
-	      if (keyptr < (keyword + sizeof(keyword) - 1))
-	        *keyptr++ = ch;
-            }
-	    else if (ch == '\"' && lastch != '\\' && !ccomment && !cstring)
-	    {
-	     /*
-	      * Start a C string constant...
-	      */
+	        if (keyptr < (keyword + sizeof(keyword) - 1))
+	          *keyptr++ = ch;
+	      }
+	      else if (ch == '\"' && lastch != '\\')
+	      {
+	       /*
+	        * Start a dbl-quote string constant...
+	        */
 
-	      cstring = -1;
-              attr    = ATTR_BLUE;
-	    }
-            else if (ch == '*' && lastch == '/' && !cstring &&
-	             PrettyPrint != PRETTY_SHELL)
-	    {
-	     /*
-	      * Start a C-style comment...
-	      */
+	        strState = StrBeg;
+		attr    = ATTR_BLUE;
+	      }
+	      else if (ch == '*' && lastch == '/' &&
+	               PrettyPrint != PRETTY_SHELL)
+	      {
+	       /*
+	        * Start a C-style comment...
+	        */
 
-	      ccomment = 1;
-	      attr     = ATTR_ITALIC | ATTR_GREEN;
-	    }
-	    else if (ch == '/' && lastch == '/' && !cstring &&
-	             PrettyPrint == PRETTY_CODE)
-	    {
-	     /*
-	      * Start a C++-style comment...
-	      */
+	        cmntState = SNTXstyl;
+	        attr     = ATTR_ITALIC | ATTR_GREEN;
+	      }
+	      else if (ch == '/' && lastch == '/' &&
+	               PrettyPrint == PRETTY_CODE)
+	      {
+	       /*
+	        * Start a C++-style comment...
+	        */
 
-	      attr = ATTR_ITALIC | ATTR_GREEN;
-	    }
-	    else if (ch == '#' && !cstring && PrettyPrint != PRETTY_CODE)
-	    {
-	     /*
-	      * Start a shell-style comment...
-	      */
+	        cmntState = NLstyl;
+	        attr = ATTR_ITALIC | ATTR_GREEN;
+	      }
+	      else if (ch == '#' && PrettyPrint != PRETTY_CODE)
+	      {
+	       /*
+	        * Start a shell-style comment...
+	        */
 
-	      attr = ATTR_ITALIC | ATTR_GREEN;
-	    }
-	    else if (ch == '#' && column == 0 && !ccomment && !cstring &&
-	             PrettyPrint == PRETTY_CODE)
-	    {
-	     /*
-	      * Start a preprocessor command...
-	      */
+	        cmntState = NLstyl;
+	        attr = ATTR_ITALIC | ATTR_GREEN;
+	      }
+	      else if (ch == '#' && column == 0 &&
+	               PrettyPrint == PRETTY_CODE)
+	      {
+	       /*
+	        * Start a preprocessor command...
+	        */
 
-	      attr = ATTR_BOLD | ATTR_RED;
+	        attr = ATTR_BOLD | ATTR_RED;
+	      }
 	    }
           }
 
@@ -1024,7 +1091,8 @@ TextMain(const char *name,	/* I - Name of filter */
 
             if (PrettyPrint)
               Page[line][i].attr = attr;
-	    else if (ch == ' ' && Page[line][i].ch)
+
+	    if (ch == ' ' && Page[line][i].ch)
 	      ch = Page[line][i].ch;
             else if (ch == Page[line][i].ch)
               Page[line][i].attr |= ATTR_BOLD;
@@ -1045,7 +1113,7 @@ TextMain(const char *name,	/* I - Name of filter */
 
           if (PrettyPrint)
 	  {
-	    if ((ch == '{' || ch == '}') && !ccomment && !cstring &&
+	    if ((ch == '{' || ch == '}') && !cmntState && !strState &&
 	        column < ColumnWidth)
 	    {
 	     /*
@@ -1063,27 +1131,27 @@ TextMain(const char *name,	/* I - Name of filter */
 
 	      Page[line][column - 1].attr = attr;
 	    }
-	    else if (ch == '\"' && lastch != '\\' && !ccomment && cstring > 0)
+	    else if (ch == '\"' && lastch != '\\' && !cmntState && strState == StrEnd)
 	    {
 	     /*
-	      * End a C string constant...
+	      * End a dbl-quote string constant...
 	      */
 
-	      cstring = 0;
+	      strState = NoStr;
 	      attr    &= ~ATTR_BLUE;
             }
-	    else if (ch == '/' && lastch == '*' && ccomment)
+	    else if (ch == '/' && lastch == '*' && cmntState)
 	    {
 	     /*
 	      * End a C-style comment...
 	      */
 
-	      ccomment = 0;
+	      cmntState = NoCmnt;
 	      attr     &= ~(ATTR_ITALIC | ATTR_GREEN);
 	    }
 
-            if (cstring < 0)
-	      cstring = 1;
+            if (strState == StrBeg)
+	      strState = StrEnd;
 	  }
 
           column ++;
@@ -1097,6 +1165,15 @@ TextMain(const char *name,	/* I - Name of filter */
     lastch = ch;
   }
 
+  /* Do not write anything if the input file is empty */
+  if (empty)
+  {
+    fprintf(stderr, "DEBUG: Input is empty, outputting empty file.\n");
+    if (fp != stdin)
+      fclose(fp);
+    return 0;
+  }
+
  /*
   * Write any remaining page data...
   */
@@ -1110,9 +1187,14 @@ TextMain(const char *name,	/* I - Name of filter */
 
   WriteEpilogue();
 
+  if (fp != stdin)
+    fclose(fp);
+
   if (ppd != NULL)
     ppdClose(ppd);
 
+  free(Page[0]);
+  free(Page);
   return (0);
 }
 
@@ -1203,7 +1285,3 @@ getutf8(FILE *fp)	/* I - File to read from */
   }
 }
 
-
-/*
- * End of "$Id$".
- */
